-----------------------------------------------------------------------------
-- Cgilua.lua: Main configuration file.
--
-- $Archive: /CGILua_LuaNG/CGI/Cgilua/Lua/cgilua.lua $
-- $Author: Anna $
-- $Date: 9/04/99 17:16 $
-- $Revision: 14 $
-----------------------------------------------------------------------------
--
-- Copyright  1995-1999 TeCGraf, PUC-Rio. 
-- Written by Anna Magdalena Hester, Renato Ferreira Borges and
-- Roberto Ierusalimschy
-- 
-- All rights reserved. 
-- 
-- Permission is hereby granted, without written agreement and
-- without license or royalty fees, to use, copy, modify, and distribute
-- this software and its documentation for any purpose, subject to the
-- following conditions: 
-- 
-- The above copyright notice and this permission notice shall appear in
-- all copies or substantial portions of this software. 
-- 
-- The names "CGILua" and "HTML Lua" cannot be used for any modified form of
-- this software that does not originate from the authors.
-- 
-- The authors specifically disclaim any warranties, including, but not
-- limited to, the implied warranties of merchantability and fitness for a
-- particular purpose. The software provided hereunder is on an "as is"
-- basis, and the authors have no obligation to provide maintenance,
-- support, updates, enhancements, or modifications. In no event shall
-- TeCGraf, PUC-Rio, or the authors be liable to any party for direct,
-- indirect, special, incidental, or consequential damages arising out of
-- the use of this software and its documentation. 


$debug


--- The table cgilua, used as a namespace

cgilua = {}


-- moves the global symbols to the table [[cgilua]]

cgilua.error = cgilua_error
cgilua_error = nil

cgilua.conffile = cgilua_conffile
cgilua_conffile = nil

cgilua.version = cgilua_version
cgilua_version = nil

cgilua.param = cgilua_param
cgilua_param = nil

cgilua.usererrormethod = cgilua_usererrormethod
cgilua_usererrormethod = nil

cgilua.getluastack = cgilua_getluastack
cgilua_getluastack = nil

cgilua.doglobal = cgilua_doglobal
cgilua_doglobal = nil

-----------------------------------------------------------------------------
-- These symbols are only defined in non-CGI APIs

cgilua.enablesession = cgilua_enablesession
cgilua_enablesession = nil

cgilua.timestamp = cgilua_timestamp
cgilua_timestamp = nil

-----------------------------------------------------------------------------
-- These functions can be defined by C in non-CGI APIs

if cgilua_getpostdata then
  cgilua.getpostdata = cgilua_getpostdata
  cgilua_getpostdata = nil
else
  cgilua.getpostdata =
  function ()
    local content_lenght = getenv( "CONTENT_LENGTH" )
    local input = read(strrep( '.', content_lenght )) 
    return input
  end
end

if cgilua_httpheader then
  cgilua.httpheader = cgilua_httpheader
  cgilua_httpheader = nil
else
  cgilua.httpheader = 
  function ( header )
    write( header )
  end
end

if cgilua_locationheader then
  cgilua.locationheader = cgilua_locationheader
  cgilua_locationheader = nil
else
  cgilua.locationheader = 
  function ( url )
    cgilua.httpheader( 'Location: '..url..'\n\n' )
  end
end

-----------------------------------------------------------------------------
-- [[cgilua.dofile]] is used to 'do' a cgilua file,
-- generating an adequate error message in case of error.
-- @param file The file do be processed.
-----------------------------------------------------------------------------
function cgilua.dofile( file )
  if not dofile( cgilua.directory..file ) then
    cgilua.error( format("Cannot open CGILua file <CODE>%s</CODE>, check installation.",
                          cgilua.directory..file) )
  end
end

-----------------------------------------------------------------------------
-- the function [[cgilua.loadlibrary]] loads a cgilua extension:
-- loads a library from the extensions directory and calls an init function.
-- @param lib The library to be loaded.
-- @param init The function of the library to be called uppon loading.
-----------------------------------------------------------------------------
function cgilua.loadlibrary( lib, init )
  local libhandle, err = loadlib( lib, cgilua.directory )
  if not libhandle then
    cgilua.error(err)
  else
    callfromlib( libhandle, init )
  end
end

-----------------------------------------------------------------------------
-- [[cgilua.directory]] points to the cgilua script directory.

cgilua.directory = gsub( cgilua.conffile, "[^\\/:]*$", "" )


-----------------------------------------------------------------------------
-- [[cgilua.external]] keeps the names of the symbols to be removed
-- from the environment during the execution of the 'map.lua' file.

cgilua.external = { "readfrom", "writeto", "appendto", "execute",
                    "remove", "rename", "loadlib" }

-----------------------------------------------------------------------------
-- Restores the functions on [[cgilua.external]] back to the
-- environment.
-- @param funcref A table containing references of the symbols
-- removed from the environment.
-----------------------------------------------------------------------------
function cgilua.restore_external( funcrefs )
  local i = 1
  while cgilua.external[i] do
    local elem = cgilua.external[i]
    if funcrefs[ elem ] then
      ref = funcrefs[ elem ]
      dostring( elem .. " = ref" )
    end
    i = i + 1
  end
  ref = nil
end

-----------------------------------------------------------------------------
-- Removes the symbols names registered on [[cgilua.external]] from
-- the environment.
-- @return A table containing the references of the removed symbols.
-----------------------------------------------------------------------------
function cgilua.remove_external()
  local i = 1
  local funcrefs = {}
  while cgilua.external[i] do
    local elem = cgilua.external[i]
    local func = dostring( "return "..elem )
    if func then
      funcrefs[ elem ] = func
      dostring( elem .. "= nil" ) 
    end
    i = i + 1
  end
  return funcrefs
end


-----------------------------------------------------------------------------
-- The function [[cgilua.mappathinfo]] maps the environment variable
-- PATH_INFO to the physical path of the lua file to be executed (the script),
-- using the mappings defined on the table [[cgilua.map]] (file [[map.lua]]).
-----------------------------------------------------------------------------
function cgilua.mappathinfo()
  local found, i = nil, 1
  local path_info = getenv("PATH_INFO")
  local luafile

  -- requires PATH_INFO
  if not path_info then
    cgilua.error( "No CGILua script specified in url." );
  end
  while not found and cgilua.map[i] do
    if strfind( path_info, '^'..cgilua.map[i][1] ) then
      luafile = gsub( path_info, '^'..cgilua.map[i][1], cgilua.map[i][2]) 
      found = 1
    end
    i = i + 1
  end
  if not luafile then
    luafile = getenv("PATH_TRANSLATED")
    if not luafile then
      cgilua.error( "CGILua script " .. path_info .. "not found." )
    end
  end
  if cgilua.default_script and filetype( luafile ) == "directory" then
    luafile = luafile .. "/" .. cgilua.default_script
  end
  return  luafile
end

-----------------------------------------------------------------------------
-- The function [[cgilua.getuserscript]] finds the script file to be executed.
-- from values given on the command line (off-line use) or from the
-- mapping of the virtual path.
-- @return The pyshical path of the user script.
-----------------------------------------------------------------------------
function cgilua.getuserscript()
  if cgilua.param then
    return cgilua.param
  else
    return cgilua.mappathinfo()
  end
end


-----------------------------------------------------------------------------
-- Define the default values for the variables expected to be defined
-- in the file 'map.lua'

cgilua.map = { }
cgi = {}

-----------------------------------------------------------------------------
-- This function is called when a pair of values (name=value) is found on the
-- on the incomming data (QUERY_STRING or POST data). The default
-- behavior is to copy the values to the table [[cgi]].
-- @param name Name of the variable.
-- @param value Value of the variable.
-----------------------------------------------------------------------------
function cgilua.insertfield( name, value )
  if not cgi[name] then
    cgi[name] = value
  else
    cgi[name] = cgi[name]..','..value
  end
end

-----------------------------------------------------------------------------
-- [[cgilua.domap]] executes the file 'map.lua', which defines some
-- configuration variables, all of them fields
-- of [[cgilua]], finds the user script to be executed, and changes
-- the process uid. this file is executed without some global symbols 
-- from iolib for security reasons; the symbols are restored 
-- only after the process uid is changed to the scripts's owner uid.
-- this function defines [[cgilua.script_path]] as the physical path
-- of the user script to be executed.
-----------------------------------------------------------------------------
function domap()
  local funcrefs = cgilua.remove_external()
  domap = nil
  cgilua.dofile( "map.lua" )
  cgilua.script_path = cgilua.getuserscript()
  -- get rid of root permissions
  owneruid( cgilua.script_path )
  cgilua.restore_external( funcrefs )
end


domap()


-----------------------------------------------------------------------------
-- The function [[splitpath]] takes a path and returns its absolute
-- directory and file parts.
-- @param path The path name to be splited.
-- @return The directory of the file and the filename.
-----------------------------------------------------------------------------
function cgilua.splitpath( path )
  local i, f, dir, file = strfind( path, "(.*[:/\\])([^:/\\]*)$" )
  if not i then
    return getcwd()..'/', path
  elseif not strfind( ':/\\', strsub(dir, 1, 1)) and
         not strfind( strsub(dir, 1, 2), '%a:')  then
    return getcwd()..'/'..dir, file
  end
  return dir, file
end
cgilua.split = cgilua.splitpath

-----------------------------------------------------------------------------
-- The function [[strsplit]] takes a string and a token
-- and splits the string on the points where the token is found.
-- @param str The string to be splited.
-- @return An array with the splited parts on the position 1,2,... and
-- the number of the elements of the array.
-----------------------------------------------------------------------------
function cgilua.strsplit( str, c )
   local t = { n=0 }

   if not str then return nil end
   gsub( str, "["..c.."]*([^"..c.."]+)["..c.."]*", 
     function ( value ) 
       %t.n = %t.n+1 
       %t[%t.n] = value 
     end )
   return t, t.n
end

-- redefines [[cgilua.directory]] to have just an absolute path

cgilua.directory = cgilua.splitpath( cgilua.conffile )



-----------------------------------------------------------------------------
-- The function [[cgilua.doscript]] process a script file,
-- executing or preprocessing it, according to the handlers defined
-- on the lua table [[cgilua.script_handlers]] ([[map.lua]] file).
-- @param script The file to be processed, with a physical path or relative
-- path (but not a URL).
-- @return Returns [[1]] in case of success or [[nil]] if no handler
-- could be found to process the file.
-----------------------------------------------------------------------------
function cgilua.doscript( script )
  local lower = strlower( script )
  local i=1
  local nomatch=1

    while cgilua.script_handlers[i] and nomatch do  
    if strfind( gsub( lower, ".*/([^/]*)", "%1" ),  cgilua.script_handlers[i].mask ) then
            
      if cgilua.script_handlers[i].handler then
        if cgilua.script_handlers[i].headerbuilder then
          local headerbuilder = dostring( "return ".. cgilua.script_handlers[i].headerbuilder )
          headerbuilder( script )
        end

        local handler = dostring( "return ".. cgilua.script_handlers[i].handler )
        handler( script )
        nomatch = nil
      else
        cgilua.error( format( "Error with handler of the file <code>%s</code>.", script ) )
      end
    end
    i = i + 1
  end
  
  if nomatch then
    return nil
  end

  return 1
end

-----------------------------------------------------------------------------
-- The function [[cgilua.responseheader]] writes the HTTP header
-- that a script may require, according to the extension
-- of the script. The table cgilua.script_handler is inquired
-- to obtain the corresponding function.
-- @param script The script being executed.
-----------------------------------------------------------------------------
function cgilua.responseheader( script )
  local lower = strlower( script )
  local i=1
  
  while cgilua.script_handlers[i] and nomatch do  
    if strfind( gsub( lower, ".*/([^/]*)", "%1" ),  cgilua.script_handlers[i].mask ) then
      local headerbuilder = dostring( "return ".. cgilua.script_handlers[i].headerbuilder )
      if headerbuilder then
        headerbuilder()
      end
    end
    i = i + 1
  end
end

-----------------------------------------------------------------------------
-- The function [[cgilua.run]] 
-- define the variables [[cgilua.cgilua_url]] [[cgilua.script_path]]
-- [[cgilua.script_vdir]] and [[cgilua.script_pdir]]
-- change the current directory to the script directory and do
-- (execute or preprocess) the script.
-----------------------------------------------------------------------------
function cgilua.run()

  if cgilua.param then

    local script_name = getenv( "SCRIPT_NAME" )
    cgilua.cgilua_url = ''
    if script_name then
      cgilua.script_vdir = cgilua.splitpath( script_name )
    end

  else

    cgilua.script_vdir  = cgilua.splitpath( getenv("PATH_INFO") )
    cgilua.cgilua_url   = getenv( "SCRIPT_NAME" )

  end

  -- extracts the directory part of the path
  cgilua.script_pdir  = cgilua.splitpath( cgilua.script_path )

  -- change directory to where the lua file resides
  chdir( cgilua.script_pdir )

  -- writes the HTTP header needed
  cgilua.responseheader( cgilua.script_path )
  
  -- do (execute ou preprocess) the script file
  if not cgilua.doscript( cgilua.script_path ) then
    cgilua.locationheader( format('http://%s%s', getenv('SERVER_NAME'), getenv('PATH_INFO')) ) 
    exit()
  end
end

-----------------------------------------------------------------------------
-- The function [[cgilua.relativeurl]] returns the url
-- to be used to activate a script in the same directory
-- of the current being executed.
-- this prevents the coding of fixed paths in the scripts,
-- keeping the scripts independent from the directory it is installed
-- (and from the directory cgilua is installed).
-- Now deprecated, use [[cgilua.mkurl]].
-----------------------------------------------------------------------------
function cgilua.relativeurl( script )
  return cgilua.cgilua_url..cgilua.script_vdir..script
end


-----------------------------------------------------------------------------
-- The global variable [[cgilua.stdin]] holds all the
-- input submited via standard input.

cgilua.stdin = nil


-----------------------------------------------------------------------------
-- The table [[cgilua.hexa]] maps hexadecimal digits
-- to its equivalent number. it is used
-- to map a string with a hexadecimal number
-- to a number.

cgilua.hexa = { h0=0,  h1=1,  h2=2,  h3=3,
                h4=4,  h5=5,  h6=6,  h7=7,
                h8=8,  h9=9,  hA=10, hB=11,
                hC=12, hD=13, hE=14, hF=15 }


-----------------------------------------------------------------------------
-- The function [[cgilua.hexa2char]] takes two hexadecimal digits
-- as strings and returns the correnpondent 
-- character with decimal code [[ms]]16+[[ls]].
-- @param ms Most significant digit.
-- @param ls Less significant digit.
-- @return The ascii caracter equivalent to the hexadecimal [[msls]]
-----------------------------------------------------------------------------
function cgilua.hexa2char( ms, ls )
  return format( "%c", cgilua.hexa['h'..ms]*16 + cgilua.hexa['h'..ls] )
end


-----------------------------------------------------------------------------
-- The function [[cgilua.unescape]] does the decoding of the
-- submited data, replacing every occurence of the
-- substring %XX (where XX is a hexadecimal number)
-- with the equivalent character of decimal code XX
-- @param str String with escaped caracters.
-- @return The unescaped string.
-----------------------------------------------------------------------------
function cgilua.unescape( str )
  str = gsub( str, "+", " " )
  str = gsub( str, "%%0D%%0A", "\n" )
  return gsub( str, "%%([0-9A-F])([0-9A-F])", cgilua.hexa2char )
end

-----------------------------------------------------------------------------
-- The function [[cgilua.parse]] parses the form data.
-- it takes the encoded data from the standard input
-- or from the environment variable QUERY_STRING,
-- depending on the form method (POST or GET).
-- it transforms the '+' character into a space (' '),
-- a sequence '%0D%0A' into an linefeed ('\n')
-- and calls the function [[pair]] for each form element encoded.
-----------------------------------------------------------------------------
function cgilua.parse()
  local input

  if getenv('REQUEST_METHOD') == 'POST' then

    if cgilua.upl_is_multipart and cgilua.upl_is_multipart() then
      cgilua.upl_form_upload()
    else
      input = cgilua.getpostdata()
      gsub( input, "&*([^&=]*)=([^&=]*)&*", function ( key, value ) cgilua.insertfield( cgilua.unescape(key), cgilua.unescape(value) ) end )
    end

  else

    input = getenv("QUERY_STRING") 
    if not input then return end
    gsub( input, "&*([^&=]*)=([^&=]*)&*", function ( key, value ) cgilua.insertfield( cgilua.unescape(key), cgilua.unescape(value) ) end )

  end

  cgilua.stdin = input
  cgilua.restorestate()
end

-----------------------------------------------------------------------------
-- The function [[cgilua.char2hexa]] converts a character
-- into a hexadecimal representation of its code number,
-- returning a string of the form %XX
-- @param c The caracter which the hexadecimal representation is being
-- inquired.
-- @return A string representing the hexadecimal representation of
-- the code number of the caracter, prefixed of '%'.
-----------------------------------------------------------------------------
function cgilua.char2hexa( c )
  return '%'..format( "%02X", ascii(c) )
end

-----------------------------------------------------------------------------
-- The function [[cgilua.escape]] does the encoding of the
-- string given by parameter, replacing every occurence 
-- of a character not in the range a-z, A-Z and 0-9
-- with a string on the form %XX, where XX is
-- the hexadecimal representation of the character's
-- number code.
-- @param s The string to be encoded.
-- @return The resulting encoded string.
-----------------------------------------------------------------------------
function cgilua.escape( s )
  s = gsub( s, "([^a-zA-Z0-9 \n])", cgilua.char2hexa )
  s = gsub( s, " ", "+" )
  s = gsub( s, "\n", "%%0D%%0A" )
  return s
end

-----------------------------------------------------------------------------
-- The function [[cgilua.encodecgi]] returns a string with all 
-- the elements of the table cgi encoded in a way
-- adequate to be used with the GET method.  
-- @return Table [[cgi]] "URL encoded".
-----------------------------------------------------------------------------
function cgilua.encodecgi()
  local str = ''
  local key,value = next( cgi, nil )

  while key do
    str = str..cgilua.escape( key )..'='..cgilua.escape( value )..'&' 
    key,value = next( cgi, key )
  end

  -- remove last &
  return strsub( str, 1, strlen(str)-1 )
end

-----------------------------------------------------------------------------
-- The function [[cgilua.encodetable]] returns a string with all 
-- the elements of the table passed as an argument encoded
-- in a way adequate to be used with the GET method.  
-- The table must be a simple associative array, without nested elements.
-- @param table The lua table which elements will be linearized.
-- @return A string "URL encoded" containing all the pairs [[name=value]],
-- of the lua table used as input.
-----------------------------------------------------------------------------
function cgilua.encodetable( table )
  local str = ''
  local key, value
  
  if not table then return nil end

  key, value = next( table, nil )
  if not key then return '' end
  while key do
    str = str..cgilua.escape( key )..'='..cgilua.escape( value )..'&'
    key,value = next( table, key )
  end

  -- remove last &
  return strsub( str, 1, strlen(str)-1 )
end


-----------------------------------------------------------------------------
-- The function [[cgilua.mkurl]] creates an url to be used to access
-- the cgilua script [[script]].
-- The file [[script]] can be an absolute or relative path.
-- @param script The file to be referenced by the URL.
-- @param table Optional argument, that specifies the arguments 
-- to be passed to the script.
-- @param stateinfo Optional argument; a handle to a state environment.
-- @return The URL built by the function.
-----------------------------------------------------------------------------
function cgilua.mkurl( script, table )
  local param = ''
  local p 

  if table then
    if cgi.CGILUASTATE then table.CGILUASTATE = cgi.CGILUASTATE end
    p = cgilua.encodetable( table )
  else
    if cgi and cgi.CGILUASTATE then
      p = cgilua.encodetable{ CGILUASTATE = cgi.CGILUASTATE }
    end
  end

  if p and p ~= '' then param = '?' .. p end

  if strsub( script, 1, 1 ) == '/' then
    return cgilua.cgilua_url .. script .. param
  else 
    return cgilua.cgilua_url .. cgilua.script_vdir .. script .. param
  end
end

-----------------------------------------------------------------------------
-- The function [[cgilua.preprocess]] outputs the contents
-- of the file specified as an argument, processing any templates
-- tags included. A response header created by the default 
-- response header handler for this file type is output before
-- the file's content (see table 
-- [[cgilua.script_handlers]] on the file [[map.lua]]).
-- Tipically this is "Content-type: text/html\n\n".
-- @param file The file to be output.
-----------------------------------------------------------------------------
function cgilua.preprocess( file )
  cgilua.responseheader( file )
  cgilua.includehtml( file )
end

-----------------------------------------------------------------------------
-- The function [[cgilua.redirect]] generates an
-- http header redirecting the browser to another URL.
-- @param The filename or the complete URL of the file.
-----------------------------------------------------------------------------
function cgilua.redirect( url, args )
  if strfind( url, "^http:" ) then
    local extra = cgilua.encodetable( args )
    if not extra or extra == '' then
      cgilua.locationheader( url )
    else
      cgilua.locationheader( url..'?'..extra )
    end
  else
    cgilua.locationheader( format( 'http://%s%s', getenv('SERVER_NAME'), 
                           cgilua.mkurl( url, args )) ) 
  end
end


-- load upload library for multipart forms

cgilua.dofile( 'upload.lua' )

-- state management

cgilua.dofile( 'state.lua' )

-- parse the form data

cgilua.parse()


-- preprocess facility

cgilua.dofile( 'customtags.lua' )
cgilua.dofile( 'preprocess.lua' )


-- define the error treatment for user scripts

cgilua.dofile( 'error.lua' )


-- define the lua environment for the user script 

cgilua.dofile( 'env.lua')


-- get and 'do' the lua file or preprocess the text file!

cgilua.run()


